
/*
 * VeriSilicon hardware context
 * Copyright (C) 2020 VeriSilicon Holdings Co., Ltd. <>
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "transfer_sidedata_vpe.h"
#include "mastering_display_metadata.h"
#include "hdr_dynamic_metadata.h"

static int vpe_transfer_side_data_vui_param_from(AVFrame *dst, VpiSidedataHandle sb_handle, int poc)
{
    VpiVuiParameters *data;
    AVVuiParameters *vui_param;
    AVFrameSideData *side_data;

    data = (VpiVuiParameters *)vpi_get_sidedata_vaddr(sb_handle);
    if(!data)
        return -1;
    side_data = av_frame_new_side_data(dst, AV_FRAME_DATA_VUI_PARAMS, sizeof(AVVuiParameters));
    if (!side_data) {
        av_log(NULL, AV_LOG_ERROR, "Allocate new AVFrame side data for vui_params failed\n");
        return -1;
    }

    vui_param = (AVVuiParameters *)side_data->data;
    vui_param->sar_width = data->sar_width;
    vui_param->sar_height = data->sar_height;
    vui_param->vui_video_signal_type_present_flag = data->vui_video_signal_type_present_flag;
    vui_param->vui_video_format = data->vui_video_format;
    vui_param->vui_video_full_range = data->vui_video_full_range;
    vui_param->color_description_present_flag = data->color_description_present_flag;
    vui_param->colour_primaries = data->color_description.colour_primaries;
    vui_param->transfer_characteristics = data->color_description.transfer_characteristics;
    vui_param->matrix_coeffs = data->color_description.matrix_coeffs;

    av_dict_set_int(&side_data->metadata, "frameId", poc , 0);
    return 0;
}

static int vpe_transfer_side_data_vui_param_to(VpiProcHandle handle, VpiFrm *dst, AVFrameSideData *side_data)
{
    VpiVuiParameters *data;
    AVVuiParameters *vui_param;
    VpiSidedataHandle sb_handle;
    VpiSidedataPoolManagerHandle pool_mgr = vpi_get_sidedata_pool_mgr(handle);

    vui_param = (AVVuiParameters *)side_data->data;
    if(!vui_param)
        return -1;
    vpi_sidedata_alloc(pool_mgr, VPI_SB_VUI_PARAMS, sizeof(VpiVuiParameters), RC_SIDE_EN, &sb_handle);
    if(!sb_handle)
        return -1;
    data = (VpiVuiParameters *)vpi_get_sidedata_vaddr(sb_handle);
    if(!data)
        return -1;
    data->sar_width = vui_param->sar_width;
    data->sar_height = vui_param->sar_height;
    data->vui_video_signal_type_present_flag = vui_param->vui_video_signal_type_present_flag;
    data->vui_video_format = vui_param->vui_video_format;
    data->vui_video_full_range = vui_param->vui_video_full_range;
    data->color_description_present_flag = vui_param->color_description_present_flag;
    data->color_description.colour_primaries = vui_param->colour_primaries;
    data->color_description.transfer_characteristics = vui_param->transfer_characteristics;
    data->color_description.matrix_coeffs = vui_param->matrix_coeffs;

    return vpifrm_attach_sidedata(dst, sb_handle);
}

static int vpe_transfer_side_data_mdcv_from(AVFrame *dst, VpiSidedataHandle sb_handle, int poc)
{
    int i;
    VpiMasterDisColVol *data;
    AVMasteringDisplayMetadata *meta;

    data = (VpiMasterDisColVol *)vpi_get_sidedata_vaddr(sb_handle);
    if(!data)
        return -1;
    meta = av_mastering_display_metadata_create_side_data(dst);
    if(!meta)
        return -1;
    // display_primaries
    for(i = 0; i < 3; i++){
        meta->display_primaries[i][0]= (AVRational){data->display_primaries_x[i], 1};
        meta->display_primaries[i][1]= (AVRational){data->display_primaries_y[i], 1};
        meta->display_primaries[i][0].den = 1;
        meta->display_primaries[i][1].den = 1;
    }
    meta->has_primaries = 1;
    // white point
    meta->white_point[0]= (AVRational){data->white_point_x, 1};
    meta->white_point[1]= (AVRational){data->white_point_y, 1};
    // luminance
    meta->max_luminance= (AVRational){data->max_display_mastering_luminance, 1};
    meta->min_luminance= (AVRational){data->min_display_mastering_luminance, 1};
    meta->has_luminance = 1;

    av_dict_set_int(&dst->side_data[dst->nb_side_data-1]->metadata, "frameId", poc , 0);

    return 0;
}

static int vpe_transfer_side_data_mdcv_to(VpiProcHandle handle, VpiFrm *dst, AVFrameSideData *side_data)
{
    int i;
    VpiMasterDisColVol *data;
    VpiSidedataHandle sb_handle;
    AVMasteringDisplayMetadata *meta;
    VpiSidedataPoolManagerHandle pool_mgr = vpi_get_sidedata_pool_mgr(handle);

    meta = (AVMasteringDisplayMetadata *)side_data->data;
    if(!meta)
        return -1;
    vpi_sidedata_alloc(pool_mgr, VPI_SB_SEI_MASTERING_DISPLAY_COLOR_VOLUME, sizeof(VpiMasterDisColVol), RC_SIDE_EN, &sb_handle);
    if(!sb_handle)
        return -1;
    data = (VpiMasterDisColVol *)vpi_get_sidedata_vaddr(sb_handle);
    // display_primaries
    for(i = 0; i < 3; i++){
        data->display_primaries_x[i] = meta->display_primaries[i][0].num / meta->display_primaries[i][0].den;
        data->display_primaries_y[i] = meta->display_primaries[i][1].num / meta->display_primaries[i][1].den;
    }
    // white point
    data->white_point_x = meta->white_point[0].num / meta->white_point[0].den;
    data->white_point_y = meta->white_point[1].num / meta->white_point[1].den;
    // luminance
    data->max_display_mastering_luminance = meta->max_luminance.num / meta->max_luminance.den;
    data->min_display_mastering_luminance = meta->min_luminance.num / meta->min_luminance.den;

    data->present_flag = 1;
    return vpifrm_attach_sidedata(dst, sb_handle);
}

static int vpe_transfer_side_data_cll_from(AVFrame *dst, VpiSidedataHandle sb_handle, int poc)
{
    VpiContentLightLevel *data;
    AVContentLightMetadata *meta;

    data = (VpiContentLightLevel *)vpi_get_sidedata_vaddr(sb_handle);
    if(!data)
        return -1;
    meta = av_content_light_metadata_create_side_data(dst);
    if(!meta)
        return -1;
    meta->MaxCLL = data->max_content_light_level;
    meta->MaxFALL = data->max_pic_average_light_level;

    av_dict_set_int(&dst->side_data[dst->nb_side_data-1]->metadata, "frameId", poc , 0);

    return 0;
}

static int vpe_transfer_side_data_cll_to(VpiProcHandle handle, VpiFrm *dst, AVFrameSideData *side_data)
{
    VpiContentLightLevel *data;
    AVContentLightMetadata *meta;
    VpiSidedataHandle sb_handle;
    VpiSidedataPoolManagerHandle pool_mgr = vpi_get_sidedata_pool_mgr(handle);

    meta = (AVContentLightMetadata *)side_data->data;
    if(!meta)
        return -1;
    vpi_sidedata_alloc(pool_mgr, VPI_SB_SEI_CONTENT_LIGHT_LEVEL, sizeof(VpiContentLightLevel), RC_SIDE_EN, &sb_handle);
    if(!sb_handle)
        return -1;
    data = (VpiContentLightLevel *)vpi_get_sidedata_vaddr(sb_handle);
    if(!data)
        return -1;
    data->max_content_light_level = meta->MaxCLL;
    data->max_pic_average_light_level = meta->MaxFALL;
    data->present_flag = 1;
    return vpifrm_attach_sidedata(dst, sb_handle);
}

static int vpe_transfer_side_data_hdr10plus_from(AVFrame *dst, VpiSidedataHandle sb_handle, int poc)
{
    int w, i, j;
    VpiT35Parameters *t35_param;
    VpiHdr10PlusMeta *data;
    AVDynamicHDRPlus *hdr10plus;
    AVHDRPlusColorTransformParams *params;

    t35_param = (VpiT35Parameters *)vpi_get_sidedata_vaddr(sb_handle);
    if(!t35_param)
        return -1;
    data = &t35_param->meta.hdr10plus;

    hdr10plus = av_dynamic_hdr_plus_create_side_data(dst);
    if(!hdr10plus)
        return -1;
    params = hdr10plus->params;

    hdr10plus->itu_t_t35_country_code = data->country_code;
    hdr10plus->application_version = data->application_version;
    hdr10plus->num_windows = data->num_windows;

    for(w = 1; w < hdr10plus->num_windows; w++){
        params[w].window_upper_left_corner_x = (AVRational){data->window_upper_left_corner_x[w], 1};
        params[w].window_upper_left_corner_y = (AVRational){data->window_upper_left_corner_y[w], 1};
        params[w].window_lower_right_corner_x = (AVRational){data->window_lower_right_corner_x[w], 1};
        params[w].window_lower_right_corner_y = (AVRational){data->window_lower_right_corner_y[w], 1};
        params[w].center_of_ellipse_x = data->center_of_ellipse_x[w];
        params[w].center_of_ellipse_y = data->center_of_ellipse_y[w];
        params[w].rotation_angle = data->rotation_angle[w];
        params[w].semimajor_axis_internal_ellipse = data->semimajor_axis_internal_ellipse[w];
        params[w].semiminor_axis_external_ellipse = data->semiminor_axis_external_ellipse[w];
        params[w].semimajor_axis_external_ellipse = data->semimajor_axis_external_ellipse[w];
        params[w].overlap_process_option = data->overlap_process_option[w];
    }

    hdr10plus->targeted_system_display_maximum_luminance = (AVRational){data->targeted_system_display_maximum_luminance, 1};
    hdr10plus->targeted_system_display_actual_peak_luminance_flag = data->targeted_system_display_actual_peak_luminance_flag;
    if(hdr10plus->targeted_system_display_actual_peak_luminance_flag){
        hdr10plus->num_rows_targeted_system_display_actual_peak_luminance = data->num_rows_targeted_system_display_actual_peak_luminance;
        hdr10plus->num_cols_targeted_system_display_actual_peak_luminance = data->num_cols_targeted_system_display_actual_peak_luminance;
        for(i = 0; i < hdr10plus->num_rows_targeted_system_display_actual_peak_luminance; i++){
            for(j = 0; j < hdr10plus->num_cols_targeted_system_display_actual_peak_luminance; j++){
                hdr10plus->targeted_system_display_actual_peak_luminance[i][j]= (AVRational){data->targeted_system_display_actual_peak_luminance[i][j], 1};
            }
        }
        for(w = 0; w < hdr10plus->num_windows; w++){
            for(i = 0; i < 3; i++){
                params[w].maxscl[i]= (AVRational){data->maxscl[w][i], 1};
            }
            params[w].average_maxrgb= (AVRational){data->average_maxrgb[w], 1};
            params[w].num_distribution_maxrgb_percentiles = data->num_distribution_maxrgb_percentiles[w];
            for(i = 0; i < params[w].num_distribution_maxrgb_percentiles; i++){
                params[w].distribution_maxrgb[i].percentage = data->distribution_maxrgb_percentages[w][i];
                params[w].distribution_maxrgb[i].percentile= (AVRational){data->distribution_maxrgb_percentiles[w][i], 1};
            }
            params[w].fraction_bright_pixels= (AVRational){data->fraction_bright_pixels[w], 1};
        }
    }

    hdr10plus->mastering_display_actual_peak_luminance_flag = data->mastering_display_actual_peak_luminance_flag;
    if(hdr10plus->mastering_display_actual_peak_luminance_flag){
        hdr10plus->num_rows_mastering_display_actual_peak_luminance = data->num_rows_mastering_display_actual_peak_luminance;
        hdr10plus->num_cols_mastering_display_actual_peak_luminance = data->num_cols_mastering_display_actual_peak_luminance;
        for(i = 0; i < hdr10plus->num_rows_mastering_display_actual_peak_luminance; i++){
            for(j = 0; j < hdr10plus->num_cols_mastering_display_actual_peak_luminance; j++){
                hdr10plus->mastering_display_actual_peak_luminance[i][j]= (AVRational){data->mastering_display_actual_peak_luminance[i][j], 1};
            }
        }
        for(w = 0; w < hdr10plus->num_windows; w++){
            params[w].tone_mapping_flag = data->tone_mapping_flag[w];
            if(params[w].tone_mapping_flag){
                params[w].knee_point_x= (AVRational){data->knee_point_x[w], 1};
                params[w].knee_point_y= (AVRational){data->knee_point_y[w], 1};
                params[w].num_bezier_curve_anchors = data->num_bezier_curve_anchors[w];
                for(i = 0; i < params[w].num_bezier_curve_anchors; i++){
                    params[w].bezier_curve_anchors[i] = (AVRational){data->bezier_curve_anchors[w][i], 1};
                }
            }
            params[w].color_saturation_mapping_flag = data->color_saturation_mapping_flag[w];
            if(params[w].color_saturation_mapping_flag){
                params[w].color_saturation_weight= (AVRational){data->color_saturation_weight[w], 1};
            }
        }
    }
    av_dict_set_int(&dst->side_data[dst->nb_side_data-1]->metadata, "frameId", poc, 0);
    return 0;
}

static int vpe_transfer_side_data_hdr10plus_to(VpiProcHandle handle, VpiFrm *dst, AVFrameSideData *side_data){
    int w, i, j;
    VpiHdr10PlusMeta *data;
    VpiT35Parameters *t35_param;
    AVDynamicHDRPlus *hdr10plus;
    VpiSidedataHandle sb_handle;
    AVHDRPlusColorTransformParams *params;
    VpiSidedataPoolManagerHandle pool_mgr = vpi_get_sidedata_pool_mgr(handle);

    hdr10plus = (AVDynamicHDRPlus *)side_data->data;
    if(!hdr10plus)
        return -1;
    params = hdr10plus->params;

    vpi_sidedata_alloc(pool_mgr, VPI_SB_SEI_HDR10_PLUS, sizeof(VpiT35Parameters), RC_SIDE_EN, &sb_handle);
    if(!sb_handle)
        return -1;
    t35_param = (VpiT35Parameters *)vpi_get_sidedata_vaddr(sb_handle);
    if(!t35_param)
        return -1;
    t35_param->meta_type = VPE_DYNAMIC_HDR10_PLUS;
    data = &t35_param->meta.hdr10plus;

    data->country_code = hdr10plus->itu_t_t35_country_code;
    data->application_version =hdr10plus->application_version;
    data->num_windows = hdr10plus->num_windows;

    for(w = 1; w < data->num_windows; w++){
        data->window_upper_left_corner_x[w] = params[w].window_upper_left_corner_x.num / params[w].window_upper_left_corner_x.den;
        data->window_upper_left_corner_y[w] = params[w].window_upper_left_corner_y.num / params[w].window_upper_left_corner_y.den;
        data->window_lower_right_corner_x[w] = params[w].window_lower_right_corner_x.num / params[w].window_lower_right_corner_x.den;
        data->window_lower_right_corner_y[w] = params[w].window_lower_right_corner_y.num / params[w].window_lower_right_corner_y.den;
        data->center_of_ellipse_x[w] = params[w].center_of_ellipse_x;
        data->center_of_ellipse_y[w] = params[w].center_of_ellipse_y;
        data->rotation_angle[w] = params[w].rotation_angle;
        data->semimajor_axis_internal_ellipse[w] = params[w].semimajor_axis_internal_ellipse;
        data->semiminor_axis_external_ellipse[w] = params[w].semiminor_axis_external_ellipse;
        data->semimajor_axis_external_ellipse[w] = params[w].semimajor_axis_external_ellipse;
        data->overlap_process_option[w] = params[w].overlap_process_option;
    }

    data->targeted_system_display_maximum_luminance =
        hdr10plus->targeted_system_display_maximum_luminance.num /
        hdr10plus->targeted_system_display_maximum_luminance.den;
    data->targeted_system_display_actual_peak_luminance_flag = hdr10plus->targeted_system_display_actual_peak_luminance_flag;
    if(data->targeted_system_display_actual_peak_luminance_flag){
        data->num_rows_targeted_system_display_actual_peak_luminance = hdr10plus->num_rows_targeted_system_display_actual_peak_luminance;
        data->num_cols_targeted_system_display_actual_peak_luminance = hdr10plus->num_cols_targeted_system_display_actual_peak_luminance;
        for (i = 0; i < data->num_rows_targeted_system_display_actual_peak_luminance; i++) {
            for (j = 0; j < data->num_cols_targeted_system_display_actual_peak_luminance; j++) {
                data->targeted_system_display_actual_peak_luminance[i][j] =
                    hdr10plus->targeted_system_display_actual_peak_luminance[i][j].num /
                    hdr10plus->targeted_system_display_actual_peak_luminance[i][j].den;
            }
        }
        for(w = 0; w < data->num_windows; w++){
            for(i = 0; i < 3; i++){
                data->maxscl[w][i] = params[w].maxscl[i].num / params[w].maxscl[i].den;
            }
            data->average_maxrgb[w] = params[w].average_maxrgb.num / params[w].average_maxrgb.den;
            data->num_distribution_maxrgb_percentiles[w] = params[w].num_distribution_maxrgb_percentiles;
            for(i = 0; i < data->num_distribution_maxrgb_percentiles[w]; i++){
                data->distribution_maxrgb_percentages[w][i] = params[w].distribution_maxrgb[i].percentage;
                data->distribution_maxrgb_percentiles[w][i] = params[w].distribution_maxrgb[i].percentile.num / params[w].distribution_maxrgb[i].percentile.den;
            }
            data->fraction_bright_pixels[w] = params[w].fraction_bright_pixels.num / params[w].fraction_bright_pixels.den;
        }
    }

    data->mastering_display_actual_peak_luminance_flag = hdr10plus->mastering_display_actual_peak_luminance_flag;
    if(data->mastering_display_actual_peak_luminance_flag){
        data->num_rows_mastering_display_actual_peak_luminance = hdr10plus->num_rows_mastering_display_actual_peak_luminance;
        data->num_cols_mastering_display_actual_peak_luminance = hdr10plus->num_cols_mastering_display_actual_peak_luminance;
        for(i = 0; i < data->num_rows_mastering_display_actual_peak_luminance; i++){
            for(j = 0; j < data->num_cols_mastering_display_actual_peak_luminance; j++){
                data->mastering_display_actual_peak_luminance[i][j] =
                    hdr10plus->mastering_display_actual_peak_luminance[i][j].num /
                    hdr10plus->mastering_display_actual_peak_luminance[i][j].den;
            }
        }
        for(w = 0; w < data->num_windows; w++){
            data->tone_mapping_flag[w] = params[w].tone_mapping_flag;
            if(data->tone_mapping_flag[w]){
                data->knee_point_x[w] = params[w].knee_point_x.num / params[w].knee_point_x.den;
                data->knee_point_y[w] = params[w].knee_point_y.num / params[w].knee_point_y.den;
                data->num_bezier_curve_anchors[w] = params[w].num_bezier_curve_anchors;
                for(i = 0; i < data->num_bezier_curve_anchors[w]; i++){
                    data->bezier_curve_anchors[w][i] = params[w].bezier_curve_anchors[i].num / params[w].bezier_curve_anchors[i].den;
                }
            }
            data->color_saturation_mapping_flag[w] = params[w].color_saturation_mapping_flag;
            if(data->color_saturation_mapping_flag[w]){
                data->color_saturation_weight[w] = params[w].color_saturation_weight.num / params[w].color_saturation_weight.den;
            }
        }
    }
    return vpifrm_attach_sidedata(dst, sb_handle);
}

static int vpe_transfer_side_data_raw_t35_from(AVFrame *dst, VpiSidedataHandle sb_handle, int poc){
    VpiRawT35Data *vpi_data;
    AVRawT35Data *raw_data;

    vpi_data = (VpiRawT35Data *)vpi_get_sidedata_vaddr(sb_handle);
    if(!vpi_data)
        return -1;

    raw_data = av_raw_t35_data_create_side_data(dst);
    if(!raw_data)
        return -1;

    if(vpi_data->payload_byte_length > 0) {
        raw_data->payload_byte_length = vpi_data->payload_byte_length;
        raw_data->payload_byte_data = (uint8_t *)av_malloc(raw_data->payload_byte_length);
        memcpy(raw_data->payload_byte_data, vpi_data->payload_byte_data, raw_data->payload_byte_length);
    }

    av_dict_set_int(&dst->side_data[dst->nb_side_data-1]->metadata, "frameId", poc, 0);
    return 0;
}

static int vpe_transfer_side_data_raw_t35_to(VpiProcHandle handle, VpiFrm *dst, AVFrameSideData *side_data){
    VpiRawT35Data *vpi_data;
    VpiSidedataHandle sb_handle;
    AVRawT35Data *raw_data;
    VpiSidedataPoolManagerHandle pool_mgr = vpi_get_sidedata_pool_mgr(handle);

    raw_data = (AVRawT35Data *)side_data->data;
    if(!raw_data)
        return -1;

    vpi_sidedata_alloc(pool_mgr, VPI_SB_SEI_RAW_T35_DATA, sizeof(VpiRawT35Data), RC_SIDE_EN, &sb_handle);
    if(!sb_handle)
        return -1;

    vpi_data = (VpiRawT35Data *)vpi_get_sidedata_vaddr(sb_handle);
    if(!vpi_data)
        return -1;

    vpi_data->payload_byte_length = raw_data->payload_byte_length;
    vpi_data->payload_byte_data = (uint8_t *)av_malloc(vpi_data->payload_byte_length);
    memcpy(vpi_data->payload_byte_data, raw_data->payload_byte_data, vpi_data->payload_byte_length);

    return vpifrm_attach_sidedata(dst, sb_handle);
}

static int transfer_sbtype_vpiframe_to_avframe(int type)
{
    switch (type) {
        case VPI_SB_ML_ROI:               return AV_FRAME_DATA_SB_ML_ROI;
        case VPI_SB_ML_CC:                return AV_FRAME_DATA_SB_ML_CC;
        case VPI_SB_ML_CXC:               return AV_FRAME_DATA_SB_ML_CXC;
        case VPI_SB_ML_LA_STAT:           return AV_FRAME_DATA_SB_ML_LA_STAT;
        case VPI_SB_ML_BOUNDING_BOX:      return AV_FRAME_DATA_SB_ML_BOUNDING_BOX;
        case VPI_SB_ML_CLASSIFICATION:    return AV_FRAME_DATA_SB_ML_CLASSIFICATION;
        case VPI_SB_ML_YDATA:             return AV_FRAME_DATA_SB_ML_YDATA;
        case VPI_SB_ML_TENSOR:            return AV_FRAME_DATA_SB_ML_TENSOR;
        default:
            return -1;
    }
}

static int vpe_transfer_side_data_ml_sb_from(AVFrame *dst, VpiSidedataHandle sb_handle, int poc){
    int sb_size, sb_type;
    AVFrameSideData *side_data = NULL;
    sb_type   = vpi_get_sidedata_type(sb_handle);
    sb_type   = transfer_sbtype_vpiframe_to_avframe(sb_type);
    sb_size   = vpi_get_sidedata_size(sb_handle);
    side_data = av_frame_new_side_data(dst, sb_type, sb_size);
    if (!side_data)
        av_log(NULL, AV_LOG_ERROR, "Add a new side data to a frame, sidedata is null.\n");

    memcpy(side_data->data, vpi_get_sidedata_vaddr(sb_handle), sb_size);

    av_dict_set_int(&side_data->metadata, "frameId", poc, 0);
    av_dict_set_int(&side_data->metadata, "width",      vpi_get_sidedata_width(sb_handle),      0);
    av_dict_set_int(&side_data->metadata, "height",     vpi_get_sidedata_height(sb_handle),     0);
    av_dict_set_int(&side_data->metadata, "channel",    vpi_get_sidedata_layoutch(sb_handle),   0);
    av_dict_set_int(&side_data->metadata, "layout",     vpi_get_sidedata_layout(sb_handle),     0);
    av_dict_set_int(&side_data->metadata, "dataformat", vpi_get_sidedata_dataformat(sb_handle), 0);
    av_dict_set_int(&side_data->metadata, "batch",      vpi_get_sidedata_batchsize(sb_handle),  0);
    return 0;
}

int vpe_transfer_side_data_from(VpiFrm *src, AVFrame *dst)
{
    int ret = 0;
    VpiSidedataHandle sb_handle;
    int sb_type, index;

    av_log(NULL, AV_LOG_DEBUG, "src %p, src->poc %d\n", src, src->poc);

    vpifrm_get_first_sidedata_by_type(src, VPI_SB_SEI_MASTERING_DISPLAY_COLOR_VOLUME, &sb_handle);
    if(sb_handle){
        ret = vpe_transfer_side_data_mdcv_from(dst, sb_handle, src->poc);
        if(ret < 0)
            av_log(NULL, AV_LOG_ERROR, "transfer sidedata_mdcv from vpe FAILED\n");
    }

    vpifrm_get_first_sidedata_by_type(src, VPI_SB_SEI_CONTENT_LIGHT_LEVEL, &sb_handle);
    if(sb_handle){
        ret = vpe_transfer_side_data_cll_from(dst, sb_handle, src->poc);
        if(ret < 0)
            av_log(NULL, AV_LOG_ERROR, "transfer sidedata_cll from vpe FAILED\n");
    }

    vpifrm_get_first_sidedata_by_type(src, VPI_SB_SEI_HDR10_PLUS, &sb_handle);
    if(sb_handle){
        ret = vpe_transfer_side_data_hdr10plus_from(dst, sb_handle, src->poc);
        if(ret < 0)
            av_log(NULL, AV_LOG_ERROR, "transfer sidedata_hdr10+ from vpe FAILED\n");
    }

    vpifrm_get_first_sidedata_by_type(src, VPI_SB_SEI_RAW_T35_DATA, &sb_handle);
    if(sb_handle){
        ret = vpe_transfer_side_data_raw_t35_from(dst, sb_handle, src->poc);
        if(ret < 0)
            av_log(NULL, AV_LOG_ERROR, "transfer sidedata_raw_t35_data from vpe FAILED\n");
    }

    vpifrm_get_first_sidedata_by_type(src, VPI_SB_VUI_PARAMS, &sb_handle);
    if(sb_handle){
        ret = vpe_transfer_side_data_vui_param_from(dst, sb_handle, src->poc);
        if(ret < 0)
            av_log(NULL, AV_LOG_ERROR, "transfer sidedata_vui_param from vpe FAILED\n");
    }

    for (sb_type = VPI_SB_ML_ROI; sb_type <= VPI_SB_ML_MAX; sb_type++){
        vpifrm_get_first_sidedata_by_type(src, sb_type, &sb_handle);
        if (sb_handle){
            ret = vpe_transfer_side_data_ml_sb_from(dst, sb_handle, src->poc);
            if(ret < 0)
                av_log(NULL, AV_LOG_ERROR, "transfer_side_data_ml_sb from vpe FAILED\n");
        }
    }
    return ret;
}

static int transfer_sbtype_avframe_to_vpiframe(int type)
{
    switch (type) {
        case AV_FRAME_DATA_SB_ML_ROI:               return VPI_SB_ML_ROI;
        case AV_FRAME_DATA_SB_ML_CC:                return VPI_SB_ML_CC;
        case AV_FRAME_DATA_SB_ML_CXC:               return VPI_SB_ML_CXC;
        case AV_FRAME_DATA_SB_ML_LA_STAT:           return VPI_SB_ML_LA_STAT;
        case AV_FRAME_DATA_SB_ML_BOUNDING_BOX:      return VPI_SB_ML_BOUNDING_BOX;
        case AV_FRAME_DATA_SB_ML_CLASSIFICATION:    return VPI_SB_ML_CLASSIFICATION;
        case AV_FRAME_DATA_SB_ML_YDATA:             return VPI_SB_ML_YDATA;
        case AV_FRAME_DATA_SB_ML_TENSOR:            return VPI_SB_ML_TENSOR;
        default:
            return VPI_SB_NONE;
    }
}

static int vpe_transfer_side_data_ml_sb_to(VpiProcHandle handle, VpiFrm *dst, AVFrameSideData *side_data){
    VpiSidedataHandle sb_handle;
    int sb_size = side_data->size;
    int sb_type = transfer_sbtype_avframe_to_vpiframe(side_data->type);
    int layout = VPE_LAYOUT_NCHW, width = 1, height = 1, chan = 1, batchsize = 1, data_format = 0;
    int nchw[4];
    AVDictionaryEntry *t = NULL;
    VpiSidedataPoolManagerHandle pool_mgr = vpi_get_sidedata_pool_mgr(handle);

    vpi_sidedata_alloc(pool_mgr, sb_type, sb_size, RC_SIDE_EN, &sb_handle);
    if (!sb_handle)
        return -1;
    memcpy(vpi_get_sidedata_vaddr(sb_handle), side_data->data, sb_size);
    t = av_dict_get(side_data->metadata, "width", NULL, 0);
    if (t)
        width = atoi(t->value);
    t = av_dict_get(side_data->metadata, "height", NULL, 0);
    if (t)
        height = atoi(t->value);
    t = av_dict_get(side_data->metadata, "layout", NULL, 0);
    if (t)
        layout = atoi(t->value);
    t = av_dict_get(side_data->metadata, "channel", NULL, 0);
    if (t)
        chan = atoi(t->value);
    t = av_dict_get(side_data->metadata, "batchsize", NULL, 0);
    if (t)
        batchsize = atoi(t->value);
    t = av_dict_get(side_data->metadata, "dataformat", NULL, 0);
    if (t)
        data_format = atoi(t->value);
    switch (layout) {
        case VPE_LAYOUT_NCHW:
            nchw[0] = width;
            nchw[1] = height;
            nchw[2] = chan;
            nchw[3] = batchsize;
            break;
        case VPE_LAYOUT_NHWC:
            nchw[0] = chan;
            nchw[1] = width;
            nchw[2] = height;
            nchw[3] = batchsize;
            break;
        case VPE_LAYOUT_CHWN:
            nchw[0] = batchsize;
            nchw[1] = width;
            nchw[2] = height;
            nchw[3] = chan;
            break;
        default:
            nchw[0] = width;
            nchw[1] = height;
            nchw[2] = chan;
            nchw[3] = batchsize;
            break;
    }
    vpi_update_side_data_attribute(sb_handle, sb_type, nchw, layout, data_format);

    return vpifrm_attach_sidedata(dst, sb_handle);
}

int vpe_transfer_side_data_to(VpiProcHandle handle, const AVFrame *src, VpiFrm *dst)
{
    int i, ret = 0;
    AVFrameSideData *side_data;
    VpiSidedataHandle sb_handle = NULL;
    int sb_type;

    for(i = 0; i < src->nb_side_data; i++){
        side_data = src->side_data[i];

        switch (side_data->type)
        {
        case AV_FRAME_DATA_MASTERING_DISPLAY_METADATA:
            ret = vpe_transfer_side_data_mdcv_to(handle, dst, side_data);
            break;
        case AV_FRAME_DATA_CONTENT_LIGHT_LEVEL:
            ret = vpe_transfer_side_data_cll_to(handle, dst, side_data);
            break;
        case AV_FRAME_DATA_DYNAMIC_HDR_PLUS:
            ret = vpe_transfer_side_data_hdr10plus_to(handle, dst, side_data);
            break;
        case AV_FRAME_DATA_RAW_T35:
            ret = vpe_transfer_side_data_raw_t35_to(handle, dst, side_data);
            break;
        case AV_FRAME_DATA_VUI_PARAMS:
            ret = vpe_transfer_side_data_vui_param_to(handle, dst, side_data);
            break;
        case AV_FRAME_DATA_SB_ML_ROI:
        case AV_FRAME_DATA_SB_ML_CC:
        case AV_FRAME_DATA_SB_ML_CXC:
        case AV_FRAME_DATA_SB_ML_LA_STAT:
        case AV_FRAME_DATA_SB_ML_BOUNDING_BOX:
        case AV_FRAME_DATA_SB_ML_CLASSIFICATION:
        case AV_FRAME_DATA_SB_ML_YDATA:
        case AV_FRAME_DATA_SB_ML_TENSOR:
            ret = vpe_transfer_side_data_ml_sb_to(handle, dst, side_data);
            break;
        default:
            break;
        }
    }
    for (sb_type = 0; sb_type < VPI_SB_TYPE_MAX; sb_type++){
        vpifrm_get_first_sidedata_by_type(dst, sb_type, &sb_handle);
        if (sb_handle){
            av_log(NULL, AV_LOG_INFO, "vpe_transfer_side_data_to has refcount! %d\n", vpi_sidedata_get_refcount(sb_handle));
            break;
        }
    }

    av_frame_remove_side_data(src, AV_FRAME_DATA_MASTERING_DISPLAY_METADATA);
    av_frame_remove_side_data(src, AV_FRAME_DATA_CONTENT_LIGHT_LEVEL);
    av_frame_remove_side_data(src, AV_FRAME_DATA_DYNAMIC_HDR_PLUS);
    av_frame_remove_side_data(src, AV_FRAME_DATA_RAW_T35);
    av_frame_remove_side_data(src, AV_FRAME_DATA_VUI_PARAMS);

    av_frame_remove_side_data(src, AV_FRAME_DATA_SB_ML_ROI);
    av_frame_remove_side_data(src, AV_FRAME_DATA_SB_ML_CC);
    av_frame_remove_side_data(src, AV_FRAME_DATA_SB_ML_CXC);
    av_frame_remove_side_data(src, AV_FRAME_DATA_SB_ML_LA_STAT);
    av_frame_remove_side_data(src, AV_FRAME_DATA_SB_ML_BOUNDING_BOX);
    av_frame_remove_side_data(src, AV_FRAME_DATA_SB_ML_CLASSIFICATION);
    av_frame_remove_side_data(src, AV_FRAME_DATA_SB_ML_YDATA);
    av_frame_remove_side_data(src, AV_FRAME_DATA_SB_ML_TENSOR);

    return ret;
}
